import copy
import datetime
import re
from collections import deque
from typing import Optional, Any


def normalize_depenv(deployment_environment: str) -> str:
    if deployment_environment.lower() == 'stable':
        return 'production'
    if deployment_environment.lower() == 'prod':
        return 'production'
    if deployment_environment.lower() == 'production':
        return 'production'
    if deployment_environment.lower() == 'staging':
        return 'staging'
    if deployment_environment.lower() == 'dev':
        return 'staging'
    if deployment_environment.lower() == 'development':
        return 'staging'
    return deployment_environment


def datetime_now(zulu=False, no_milliseconds=True) -> datetime.datetime:
    res = datetime.datetime.now(datetime.timezone.utc)
    if not zulu:
        res = res.astimezone(datetime.timezone(datetime.timedelta(hours=0), name="UTC"))
    if no_milliseconds:
        res = res.replace(microsecond=0)
    assert res.tzinfo is not None  # enforce timezone aware (non-naive) datetime
    return res


def any_to_opt_bool(value: Any, default: Optional[bool] = None) -> Optional[bool]:
    if value is None:
        return default
    if value == 1:
        return True
    if value == 0:
        return False
    if str(value).lower() in ['true', 't', '1', 'yes']:
        return True
    if str(value).lower() in ['false', 'f', '0', 'no']:
        return False
    return default


def any_to_bool(value: Any, default: Optional[bool] = None) -> bool:
    res = any_to_opt_bool(value, default)
    if res is not None:
        return res
    if default is None:
        raise ValueError('Could not convert "{}" to boolean value'.format(value))
    return default


def duration_string_to_seconds(duration: str) -> Optional[int]:
    if not duration:
        return None
    match = re.match(r'^([0-9.]+) *([a-zA-Z]+)$', duration)
    if not match:
        return None

    num = float(match.group(1)) if '.' in match.group(1) else int(match.group(1))
    unit = match.group(2).lower()
    if unit == 'h' or unit == 'hour' or unit == 'hours':
        return int(num * 60 * 60)
    elif unit == 'm' or unit == 'min' or unit == 'mins' or unit == 'minute' or unit == 'minutes':
        return int(num * 60)
    elif unit == 's' or unit == 'sec' or unit == 'secs' or unit == 'second' or unit == 'seconds':
        return int(num)
    elif unit == 'day' or unit == 'days':
        return int(num * 60 * 60 * 24)
    elif unit == 'week' or unit == 'weeks':
        return int(num * 60 * 60 * 24 * 7)
    else:
        return None
    # Impossible to support because not fixed length: month, year, ...


def deep_update_dict(target: dict, extra: dict) -> dict:
    """
    Deep update by adding all keys (nested) of extra to target.
    Nested arrays are copied from extra to target with copy.deepcopy(), so no deep_update is done on them!

    Notes:
       - this deeply modifies target, so make sure you first copy.deepcopy it if needed!
       - you can't remove keys from target with this, only add them
       - type conflicts between values of target and extra are handled by extra overriding target

    :param target: the dict that should be updated deeply
    :param extra: extra keys to be added (nested)
    :return: target
    """

    # deque to avoid recursion
    jobs = deque()
    jobs.append( (target, extra) )

    while len(jobs) > 0:
        job_target, job_extra = jobs.popleft()

        for key,extra_value in job_extra.items():
            if isinstance(extra_value, dict):
                if key in job_target and isinstance(job_target[key], dict):
                    jobs.append( (job_target[key], extra_value) )
                else:
                    job_target[key] = copy.deepcopy(extra_value)
            else:
                job_target[key] = extra_value

    return target


def strip_null_from_json_dict(json_dict: dict, *,
                              strip_empty_dict: bool = False,
                              strip_empty_list: bool = False,
                              process_lists: bool = False,
                              ):
    """
    Strip null values from dicts, also stripping empty dicts and list values if requested.
    Doesn't look inside lists unless process_lists=True, so doesn't:
       - Strip dicts inside lists.
       - Remove empty dicts inside lists.

    (recursive-less implementation, which adds a bit of complexity.)

    :param json_dict: The original json dict. Will not be modified.
    :param strip_empty_dict: remove empty dicts
    :param strip_empty_list: remove empty lists
    :param process_lists: modify nested lists
    :return:
    """
    res = copy.deepcopy(json_dict)

    todo = deque()
    todo.append((res, []))

    while len(todo) > 0:
        current, parents = todo.popleft()
        current_is_dict = isinstance(current, dict)

        if current_is_dict:
            empty = []
            for key, value in current.items():
                if value is None:
                    empty.append(key)
                elif isinstance(value, dict):
                    if len(value) > 0:
                        todo.append((value, parents + [current]))
                    if len(value) == 0 and strip_empty_dict:
                        empty.append(key)
                elif isinstance(value, list):
                    if len(value) > 0 and process_lists:
                        todo.append((value, parents + [current]))
                    if len(value) == 0 and strip_empty_list:
                        empty.append(key)
            for key in empty:
                del current[key]
        else:
            assert isinstance(current, list)
            empty_idx = []
            for i, value in enumerate(current):
                if value is None:
                    empty_idx.append(i)
                elif isinstance(value, dict):
                    if len(value) > 0:
                        todo.append((value, parents + [current]))
                    if len(value) == 0 and strip_empty_dict:
                        empty_idx.append(i)
                elif isinstance(value, list):
                    if len(value) > 0 and process_lists:
                        todo.append((value, parents + [current]))
                    if len(value) == 0 and strip_empty_list:
                        empty_idx.append(i)
            for i in reversed(empty_idx):
                del current[i]

        if len(current) == 0 and parents and (strip_empty_dict if current_is_dict else strip_empty_list):
            parent_parents = list(parents)
            parent = parent_parents.pop()
            todo.append((parent, parent_parents))
    return res



# ----------------------------------------------------------------------
# Copyright (c) 2010-2016 Raytheon BBN Technologies
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and/or hardware specification (the "Work") to
# deal in the Work without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Work, and to permit persons to whom the Work
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Work.
#
# THE WORK IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE WORK OR THE USE OR OTHER DEALINGS
# IN THE WORK.
# ----------------------------------------------------------------------
'''
URN creation and verification utilities.
'''

import re
from functools import total_ordering
from typing import Union, Optional

URN_PREFIX = "urn:publicid:IDN"

@total_ordering
class URN(object):
    """
    A class that creates and extracts values from URNs
    URN Convention:
    urn:publicid:IDN+<authority>+<type>+<name>
    Authority, type, and name are public ids transcribed into URN format
    By convention a CH's name should be "ch" and an AM's should be "am"
    The authority of the CH should be the prefix for all of your AM and user authorities
    For instance: CH authority = "gcf//gpo//bbn", AM authority = "gcf//gpo/bbn//am1", user authority = "gcf//gpo//bbn"

    EXAMPLES:

    ch_urn = URN("gcf//gpo//bbn", "authority", "sa").urn_string() for a clearinghouse URN
    am1_urn = URN("gcf//gpo//bbn//site1", "authority", "am").urn_string() for an AM at this authority
        Looks like urn:publicid:IDN+gcf:gpo:bbn:site1+authority+am
    am2_urn = URN("gcf//gpo//bbn//site2", "authority", "am").urn_string() for a second AM at this authority
        Looks like urn:publicid:IDN+gcf:gpo:bbn:site2+authority+am
    user_urn = URN("gcf//gpo//bbn", "user", "jane").urn_string() for a user made by the clearinghouse
        Looks like urn:publicid:IDN+gcf:gpo:bbn+user+jane
    slice_urn = URN("gcf//gpo//bbn", "slice", "my-great-experiment").urn_string()
        Looks like urn:publicid:IDN+gcf:gpo:bbn+slice+my-great-experiment
    resource_at_am1_urn = URN("gcf//gpo//bbn/site1", "node", "LinuxBox23").urn_string() for Linux Machine 23 managed by AM1 (at site 1)
        Looks like urn:publicid:IDN+gcf:gpo:bbn:site1+node+LinuxBox23
    """

    def __init__(self, urn=None, *, authority=None, type=None, name=None):
        if not urn is None:
            if not is_valid_urn(urn):
                raise ValueError("Invalid URN %s" % urn)

            spl = urn.split('+')
            if len(spl) < 4:
                raise ValueError("Invalid URN %s" % urn)
            self.authority = urn_to_string_format(spl[1])
            self.type = urn_to_string_format(spl[2])
            self.name = urn_to_string_format('+'.join(spl[3:]))
            self.urn = urn
        else:
            if not authority or not type or not name:
                raise ValueError("Must provide either all of authority, type, and name, or a urn must be provided")

            for i in [authority, type, name]:
                if i.strip() == '':
                    raise ValueError("Parameter to create_urn was empty string")

            self.authority = authority
            self.type = type
            self.name = name

            # FIXME: check these are valid more?
            if not is_valid_urn_string(authority):
                authority = string_to_urn_format(authority)
            if not is_valid_urn_string(type):
                type = string_to_urn_format(type)
            if not is_valid_urn_string(name):
                name = string_to_urn_format(name)

            self.urn = '%s+%s+%s+%s' % (URN_PREFIX, authority, type, name)
            if not is_valid_urn(self.urn):
                raise ValueError(
                    "Failed to create valid URN from args %s, %s, %s" % (self.authority, self.type, self.name))

        lower_authority = string_to_urn_format(self.authority).lower()
        lower_type = string_to_urn_format(self.type).lower()
        lower_name = string_to_urn_format(self.name).lower()
        self.lowercase_urn_string = '%s+%s+%s+%s' % (URN_PREFIX, lower_authority, lower_type, lower_name)

    def __str__(self):
        return self.urn_string()

    def __repr__(self):
        return 'URN("'+self.urn_string()+'")'

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, URN):
            return self.lowercase_urn_string == other.lowercase_urn_string
        return NotImplemented

    def __lt__(self, other):
        if not isinstance(other, URN):
            return NotImplemented
        return self.lowercase_urn_string < other.lowercase_urn_string

    def __hash__(self):
        """Overrides the default implementation"""
        return hash(self.lowercase_urn_string)

    def urn_string(self):
        return self.urn

    def getAuthority(self):
        '''Get the authority in un-escaped publicid format'''
        return self.authority

    def getType(self):
        '''Get the URN type in un-escaped publicid format'''
        return self.type

    def getName(self):
        '''Get the name in un-escaped publicid format'''
        return self.name

    @property
    def authority_parts(self):
        return self.authority.split('//')


# Translate publicids to URN format.
# The order of these rules matters
# because we want to catch things like double colons before we
# translate single colons. This is only a subset of the rules.
# See the GENI Wiki: GAPI_Identifiers
# See http://www.faqs.org/rfcs/rfc3151.html
publicid_xforms = [('%', '%25'),
                   (';', '%3B'),
                   ('+', '%2B'),
                   (' ', '+'),  # note you must first collapse WS
                   ('#', '%23'),
                   ('?', '%3F'),
                   ("'", '%27'),
                   ('::', ';'),
                   (':', '%3A'),
                   ('//', ':'),
                   ('/', '%2F')]

# FIXME: See sfa/util/xrn/Xrn.URN_PREFIX which is ...:IDN
publicid_urn_prefix = 'urn:publicid:'


def always_urn(urn: Union[str, URN]) -> URN:
    """
    throws exception when urn is None or invalid URN
    :param urn: str or URN object already
    :return:
    """
    if not urn:
        raise ValueError("Invalid URN {}".format(urn))
    return urn if isinstance(urn, URN) else URN(urn=urn)


def always_optional_urn(urn: Union[str, URN]) -> Optional[URN]:
    """
    returns None when urn is None or invalid URN
    :param urn: str or URN object already
    :return:
    """
    if urn:
        if isinstance(urn, URN):
            return urn
        elif is_valid_urn(urn):
            return URN(urn=urn)
        else:
            return None
    else:
        return None


def nameFromURN(instr):
    '''Get the name from the given URN, or empty if not a valid URN'''
    if not instr:
        return ""
    try:
        urn = URN(urn=instr)
        return urn.getName()
    except Exception as e:
        #        print 'exception parsing urn: %s' % e
        return ""


# validate urn
# Note that this is not sufficient but it is necessary
def is_valid_urn_string(instr):
    '''Could this string be part of a URN'''
    if instr is None or not isinstance(instr, str):
        return False
    # No whitespace
    # no # or ? or /
    if re.search("[\s|\?\/\#]", instr) is None:
        return True
    return False


# Note that this is not sufficient but it is necessary
def is_valid_urn(inurn):
    ''' Check that this string is a valid URN'''
    # FIXME: This could pull out the type and do the type specific
    # checks that are currently below
    # FIXME: This should check for non empty authority and name pieces
    return is_valid_urn_string(inurn) and \
           inurn.startswith(publicid_urn_prefix) and \
           len(inurn.split('+')) > 3

class UrnError(Exception):
    def __init__(self, message):
        self.message = message


def is_valid_urn_bytype(inurn, urntype, logger=None):
    try:
        return check_valid_urn_bytype(inurn, urntype, logger)
    except UrnError as e:
        if logger:
            logger.warn(e.message)
        return False


def check_valid_urn_bytype(inurn, urntype, logger=None):
    if not is_valid_urn(inurn):
        raise UrnError('Does not look like an URN')
    urnObj = URN(urn=inurn)
    if not urntype:
        urntype = ""
    urntype = urntype.lower()
    if not urnObj.getType().lower() == urntype:
        raise UrnError("URN {} not of right type: {}, not {}".format(inurn, urnObj.getType().lower(), urntype))
    authname = urnObj.getAuthority()
    if len(authname) == 0:
        raise UrnError("URN {} has empty authority".format(inurn))
    if len(authname) > 253:
        raise UrnError("URN {} has too long authority".format(inurn))
    if not re.match('[a-zA-Z0-9][a-zA-Z0-9_:.-]*', authname):
        raise UrnError("URN authority may only be alphanumeric hyphen colon dot underscore "
                       "(and only leading alphanumeric): {}".format(authname))
    name = urnObj.getName()
    if urntype == 'slice':
        # Slice names are <=19 characters, only alphanumeric plus hyphen (no hyphen in first character): '^[a-zA-Z0-9][-a-zA-Z0-9]{0,18}$'
        if len(name) > 19:
            raise UrnError("URN {} too long. Slice names are max 19 characters".format(inurn))
        if not re.match("^[a-zA-Z0-9][-a-zA-Z0-9]{0,18}$", name):
            raise UrnError("Slice names may only be alphanumeric plus hyphen (no leading hyphen): {}".format(name))
    elif urntype == 'sliver':
        # May use only alphanumeric characters plus hyphen
        # Note that EG uses a ':' as well.
        if not re.match("^[-a-zA-Z0-9_.]+$", name):
            raise UrnError("Sliver names may only be alphanumeric plus hyphen, underscore, or period: {}".format(name))
    elif urntype == 'user':
        # Usernames should begin with a letter and be alphanumeric or underscores; no hyphen or '.': ('^[a-zA-Z][\w]{0,7}$').
        # Usernames are limited to 8 characters.
        if len(name) > 8:
            raise UrnError("URN {} too long. User names are max 8 characters".format(inurn))
        if not re.match("^[a-zA-Z][\w]{0,7}$", name):
            raise UrnError("User names may only be alphanumeric plus underscore, beginning with a letter: {}".format(inurn))
    elif urntype == 'project':
        # Project name rules:
        #  - Min length 2
        #  - Max length 32
        #  - First character alphabetic
        #  - Subsequent characters alphanumber plus hyphen and underscore
        PROJECT_NAME_REGEX = '^[a-zA-Z][a-zA-Z0-9-_]{1,31}$'
        if len(name) > 32:
            raise UrnError("URN {} too long. Project names are max 32 characters".format(inurn))
        if len(name) < 2:
            raise UrnError("URN {} too short. Project names are min 2 characters".format(inurn))
        if not re.match(PROJECT_NAME_REGEX, name):
            raise UrnError("Project names may only be alphanumeric plus underscore and hyphen, "
                            "beginning with a letter: {}".format(name))
    elif len(name) == 0:
        raise UrnError("Empty name in URN {}".format(inurn))
    return True


def urn_to_publicid(urn):
    '''Convert a URN like urn:publicid:... to a publicid'''
    # Remove prefix
    if urn is None or not is_valid_urn(urn):
        # Erroneous urn for conversion
        raise ValueError('Invalid urn: ' + urn)
    publicid = urn[len(publicid_urn_prefix):]
    # return the un-escaped string
    return urn_to_string_format(publicid)


def publicid_to_urn(id):
    '''Convert a publicid to a urn like urn:publicid:.....'''
    # prefix with 'urn:publicid:' and escape chars
    return publicid_urn_prefix + string_to_urn_format(id)


def string_to_urn_format(instr):
    '''Make a string URN compatible, collapsing whitespace and escaping chars'''
    if instr is None or instr.strip() == '':
        raise ValueError("Empty string cant be in a URN")
    # Collapse whitespace
    instr = ' '.join(instr.strip().split())
    for a, b in publicid_xforms:
        instr = instr.replace(a, b)
    return instr


def urn_to_string_format(urnstr):
    '''Turn a part of a URN into publicid format, undoing transforms'''
    if urnstr is None or urnstr.strip() == '':
        return urnstr
    publicid = urnstr
    # Validate it is reasonable URN string?
    for a, b in reversed(publicid_xforms):
        publicid = publicid.replace(b, a)
    return publicid



import re
from typing import List

_BASH_NO_QUOTE_PATTERN = re.compile(r'[/:.A-Za-z0-9_-]+')


def need_bash_quote(arg: str) -> bool:
    """
    returns True if the argument must be quoted when used as a command line argument on a bash shell.
    """
    return len(arg) == 0 or not _BASH_NO_QUOTE_PATTERN.fullmatch(arg)


def bash_quote(arg: str) -> str:
    """
    Quotes the argument, for bash.
    This handles quote marks correctly for bash (which looks very ugly in bash).
    """
    if need_bash_quote(arg):
        return "'"+arg.replace("\'", "'\\''")+"'"
    else:
        return arg


def bash_cmd_array_to_str(cmd: List[str]) -> str:
    """
    Convert a command array, like you pass to Popen, to a string command that can safely be executed on a shell.
    This quotes the arguments needed.
    """
    return ' '.join([bash_quote(arg) for arg in cmd])



from .sentry_utils import find_sentry_release_id

from .urn_util import URN, always_urn, always_optional_urn, check_valid_urn_bytype

from .utils import normalize_depenv, datetime_now, any_to_opt_bool, any_to_bool, duration_string_to_seconds, \
    deep_update_dict, strip_null_from_json_dict

from .validation_utils import is_valid_email, is_valid_ssh_key, is_valid_uuid4, is_valid_uuid, ALLOWED_HOST_KEY_ALGOS

__all__ = [
    find_sentry_release_id,
    URN, always_urn, always_optional_urn, check_valid_urn_bytype,
    normalize_depenv, datetime_now, any_to_opt_bool, any_to_bool, duration_string_to_seconds,
    deep_update_dict, strip_null_from_json_dict,
    is_valid_email, is_valid_ssh_key, is_valid_uuid4, is_valid_uuid, ALLOWED_HOST_KEY_ALGOS
]



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="imecilabt-utils",
    version="1.4.0",

    description="A collection of functions useful in various projects iLab.t projects.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    url="https://gitlab.ilabt.imec.be/ilabt/imecilabt-py-utils",

    author="Wim Van de Meerssche",
    author_email="wim.vandemeerssche@ugent.be",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],

    packages=["imecilabt_utils"],
    include_package_data=True,

    install_requires=[],
    python_requires='>=3.5',

    # Tests with py.test:  run with: python setup.py pytest
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    zip_safe=False,
)



from uuid import UUID
from typing import Optional


def is_valid_email(email: Optional[str]) -> bool:
    """
    Basic check if a string is an email address email is valid.
    (Could be extended to check for illegal chars etc.)
    :param email:
    :return: False if the string doesn't represent an email address. True if it probably does.
    """
    if not email:
        return False
    email = str(email.strip())
    if email.count('@') == 1 and '.' in email and len(email) >= 6:
        return True
    return False


ALLOWED_HOST_KEY_ALGOS = [
    # current secure versions according to https://github.com/jtesta/ssh-audit
    'ssh-ed25519-cert-v01@openssh.com,',
    'ssh-rsa-cert-v01@openssh.com',
    'ssh-ed25519',
    'ssh-rsa',
]


def is_valid_ssh_key(ssh_pub_key: Optional[str]) -> bool:
    """
    Checks if a string looks like an OpenSSH public key string. (= the format found in ~/.ssh/id_rsa.pub)
    Only a basic check: this does not decode the SSH public key to check.
    :param ssh_pub_key: a string to check
    :return: False if the string does not contain an OpenSSH public key string. True if it probably does.
    """
    if not ssh_pub_key:
        return False
    ssh_pub_key = str(ssh_pub_key.strip())
    key_parts = ssh_pub_key.split(' ')
    host_algo = key_parts[0]
    if host_algo in ALLOWED_HOST_KEY_ALGOS and len(key_parts) >= 2:
        return True
    return False


def is_valid_uuid4(uuid_str: str) -> bool:
    try:
        uuid_obj = UUID(uuid_str, version=4)
    except ValueError:
        return False

    return str(uuid_obj) == uuid_str


def is_valid_uuid(uuid_str: str) -> bool:
    try:
        uuid_obj = UUID(uuid_str)
    except ValueError:
        return False

    recontructed_uuid_str = str(uuid_obj)
    assert len(recontructed_uuid_str) == 36, "is_valid_uuid({}) uuid_obj={} recontructed_uuid_str={} len={}"\
        .format(uuid_str, uuid_obj, recontructed_uuid_str, len(recontructed_uuid_str) == 36)
    return recontructed_uuid_str.lower() == uuid_str.lower()



import logging
import re
from subprocess import PIPE, STDOUT, Popen
from typing import Optional
from pathlib import Path


def _call_git(*params: str, **kwargs) -> str:
    res = ''
    com = ['git']
    com.extend(params)

    kwargs['stdout'] = PIPE
    kwargs['stderr'] = STDOUT
    kwargs['universal_newlines'] = True

    # Annoyingly, git doesn't seem to listen to GIT_ASKPASS and SSH_ASKPASS, and use TTY to get pass.
    # Solution: from TTY: use "setsid -w <git command>"
    #           not from TTY: should work
    with open('/dev/null', 'r') as devnull:
        kwargs['stdin'] = devnull
        with Popen(com, **kwargs) as proc:
            for line in proc.stdout:
                res += line
            res_code = proc.wait()
            if res_code != 0:
                raise Exception('Command returned error ({}): {} (command={} )'.format(res_code, res.strip(), com))
    return res.strip()


def _call_sentry_cli(*params: str, **kwargs) -> str:
    res = ''
    com = ['sentry-cli']
    com.extend(params)

    kwargs['stdout'] = PIPE
    kwargs['stderr'] = STDOUT
    kwargs['universal_newlines'] = True

    with Popen(com, **kwargs) as proc:
        for line in proc.stdout:
            res += line
        res_code = proc.wait()
        if res_code != 0:
            raise Exception('Command returned error ({}): {} (command={} )'.format(res_code, res.strip(), com))
    return res.strip()


def _find_git_sha(repo_filename: str) -> Optional[str]:
    source_path = Path(repo_filename).resolve()
    if not source_path:
        return None
    source_dir = source_path.parent
    if not source_dir:
        return None
    try:
        res = _call_git("rev-parse", "HEAD", cwd=source_dir).strip()
        assert re.compile(r'^([0-9a-f]+)$').match(res)
        assert len(res) > 10
        return res
    except:
        logging.error('Failed "git rev-parse head" in "{}" (will ignore)'.format(source_dir), exc_info=True)
        return None


def _find_git_branch(repo_filename: str) -> Optional[str]:
    source_path = Path(repo_filename).resolve()
    if not source_path:
        return None
    source_dir = source_path.parent
    if not source_dir:
        return None
    try:
        res = _call_git("rev-parse", "--abbrev-ref", "HEAD", cwd=source_dir).strip()
        assert re.compile(r'^([^ ]+)$').match(res)
        assert len(res) > 3
        return res
    except:
        logging.error('Failed "git rev-parse --abbrev-ref HEAD" in "{}" (will ignore)'.format(source_dir), exc_info=True)
        return None


def find_sentry_release_id(repo_filename: str) -> Optional[str]:
    try:
        res = _call_sentry_cli("releases", "propose-version").strip()
        assert re.compile(r'^([0-9a-f]+)$').match(res)
        assert len(res) > 10
        logging.info('Sentry Release ID found: {}'.format(res))
        return res
    except:
        logging.info('Sentry Release ID not found using sentry-cli. Will use fallback using git.')

    # FALLBACK
    sha = _find_git_sha(repo_filename)
    # branch = _find_git_branch(repo_filename)
    if sha:
        res = sha
        logging.info('Sentry Release ID found: {}'.format(res))
        return res
    # if sha and branch:
    #     res = '{}-{}'.format(branch, sha)
    #     logging.info('Sentry Release ID found: {}'.format(res))
    #     return res
    else:
        logging.warning('Sentry Release ID NOT FOUND')
        return None